@@ -3,19 +3,32 @@ import collections
 import random
 import socket
 import warnings
-from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union
+from typing import (
+    Any,
+    Deque,
+    Dict,
+    Generator,
+    List,
+    Mapping,
+    Optional,
+    Type,
+    TypeVar,
+    Union,
+)
 
-from redis.asyncio.client import EMPTY_RESPONSE, NEVER_DECODE, AbstractRedis
+from redis.asyncio.client import ResponseCallbackT
 from redis.asyncio.connection import Connection, DefaultParser, Encoder, parse_url
 from redis.asyncio.parser import CommandsParser
+from redis.client import EMPTY_RESPONSE, NEVER_DECODE, AbstractRedis
 from redis.cluster import (
+    PIPELINE_BLOCKED_COMMANDS,
     PRIMARY,
     READ_COMMANDS,
     REPLICA,
     SLOT_ID,
     AbstractRedisCluster,
     LoadBalancer,
-    cleanup_kwargs,
+    block_pipeline_command,
     get_node_name,
     parse_cluster_slots,
 )
@@ -37,13 +50,40 @@ from redis.exceptions import (
     TimeoutError,
     TryAgainError,
 )
-from redis.typing import EncodableT, KeyT
-from redis.utils import dict_merge, str_if_bytes
+from redis.typing import AnyKeyT, EncodableT, KeyT
+from redis.utils import dict_merge, safe_str, str_if_bytes
 
 TargetNodesT = TypeVar(
-    "TargetNodesT", "ClusterNode", List["ClusterNode"], Dict[Any, "ClusterNode"]
+    "TargetNodesT", str, "ClusterNode", List["ClusterNode"], Dict[Any, "ClusterNode"]
 )
 
+CONNECTION_ALLOWED_KEYS = (
+    "client_name",
+    "db",
+    "decode_responses",
+    "encoder_class",
+    "encoding",
+    "encoding_errors",
+    "health_check_interval",
+    "parser_class",
+    "password",
+    "redis_connect_func",
+    "retry",
+    "retry_on_timeout",
+    "socket_connect_timeout",
+    "socket_keepalive",
+    "socket_keepalive_options",
+    "socket_read_size",
+    "socket_timeout",
+    "socket_type",
+    "username",
+)
+
+
+def cleanup_kwargs(**kwargs: Any) -> Dict[str, Any]:
+    """Remove unsupported or disabled keys from kwargs."""
+    return {k: v for k, v in kwargs.items() if k in CONNECTION_ALLOWED_KEYS}
+
 
 class ClusterParser(DefaultParser):
     EXCEPTION_CLASSES = dict_merge(
@@ -131,7 +171,7 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
     """
 
     @classmethod
-    def from_url(cls, url: str, **kwargs) -> "RedisCluster":
+    def from_url(cls, url: str, **kwargs: Any) -> "RedisCluster":
         """
         Return a Redis client object configured from the given URL.
 
@@ -201,7 +241,7 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
         cluster_error_retry_attempts: int = 3,
         reinitialize_steps: int = 10,
         url: Optional[str] = None,
-        **kwargs,
+        **kwargs: Any,
     ) -> None:
         if not startup_nodes:
             startup_nodes = []
@@ -212,7 +252,7 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
                 "Argument 'db' is not possible to use in cluster mode"
             )
 
-        # Get the startup node/s
+        # Get the startup node(s)
         if url:
             url_options = parse_url(url)
             if "path" in url_options:
@@ -247,34 +287,34 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
         self.connection_kwargs = kwargs = cleanup_kwargs(**kwargs)
         self.response_callbacks = kwargs[
             "response_callbacks"
-        ] = self.__class__.RESPONSE_CALLBACKS
+        ] = self.__class__.RESPONSE_CALLBACKS.copy()
         if host and port:
             startup_nodes.append(ClusterNode(host, port, **self.connection_kwargs))
 
+        self.nodes_manager = NodesManager(
+            startup_nodes=startup_nodes,
+            require_full_coverage=require_full_coverage,
+            **self.connection_kwargs,
+        )
         self.encoder = Encoder(
             kwargs.get("encoding", "utf-8"),
             kwargs.get("encoding_errors", "strict"),
             kwargs.get("decode_responses", False),
         )
         self.cluster_error_retry_attempts = cluster_error_retry_attempts
-        self.command_flags = self.__class__.COMMAND_FLAGS.copy()
-        self.node_flags = self.__class__.NODE_FLAGS.copy()
         self.read_from_replicas = read_from_replicas
-        self.reinitialize_counter = 0
         self.reinitialize_steps = reinitialize_steps
-        self.nodes_manager = NodesManager(
-            startup_nodes=startup_nodes,
-            require_full_coverage=require_full_coverage,
-            **self.connection_kwargs,
-        )
 
-        self.result_callbacks = self.__class__.RESULT_CALLBACKS
+        self.reinitialize_counter = 0
+        self.commands_parser = CommandsParser()
+        self.node_flags = self.__class__.NODE_FLAGS.copy()
+        self.command_flags = self.__class__.COMMAND_FLAGS.copy()
+        self.result_callbacks = self.__class__.RESULT_CALLBACKS.copy()
         self.result_callbacks[
             "CLUSTER SLOTS"
         ] = lambda cmd, res, **kwargs: parse_cluster_slots(
             list(res.values())[0], **kwargs
         )
-        self.commands_parser = CommandsParser()
         self._initialize = True
         self._lock = asyncio.Lock()
 
@@ -310,16 +350,14 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
     async def __aexit__(self, exc_type: None, exc_value: None, traceback: None) -> None:
         await self.close()
 
-    def __await__(self):
+    def __await__(self) -> Generator[Any, None, "RedisCluster"]:
         return self.initialize().__await__()
 
     _DEL_MESSAGE = "Unclosed RedisCluster client"
 
-    def __del__(self, _warnings=warnings):
+    def __del__(self) -> None:
         if hasattr(self, "_initialize") and not self._initialize:
-            _warnings.warn(
-                f"{self._DEL_MESSAGE} {self!r}", ResourceWarning, source=self
-            )
+            warnings.warn(f"{self._DEL_MESSAGE} {self!r}", ResourceWarning, source=self)
             try:
                 context = {"client": self, "message": self._DEL_MESSAGE}
                 # TODO: Change to get_running_loop() when dropping support for py3.6
@@ -408,7 +446,7 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
 
         self.nodes_manager.default_node = node
 
-    def set_response_callback(self, command: KeyT, callback: Callable) -> None:
+    def set_response_callback(self, command: str, callback: ResponseCallbackT) -> None:
         """Set a custom response callback."""
         self.response_callbacks[command] = callback
 
@@ -430,9 +468,8 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
         return key_slot(k)
 
     async def _determine_nodes(
-        self, *args, node_flag: Optional[str] = None
+        self, command: str, *args: Any, node_flag: Optional[str] = None
     ) -> List["ClusterNode"]:
-        command = args[0]
         if not node_flag:
             # get the nodes group for this command if it was predefined
             node_flag = self.command_flags.get(command)
@@ -457,16 +494,15 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
         # get the node that holds the key's slot
         return [
             self.nodes_manager.get_node_from_slot(
-                await self._determine_slot(*args),
+                await self._determine_slot(command, *args),
                 self.read_from_replicas and command in READ_COMMANDS,
             )
         ]
 
-    async def _determine_slot(self, *args) -> int:
-        command = args[0]
+    async def _determine_slot(self, command: str, *args: Any) -> int:
         if self.command_flags.get(command) == SLOT_ID:
             # The command contains the slot ID
-            return args[1]
+            return int(args[0])
 
         # Get the keys in the command
 
@@ -478,19 +514,17 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
         # - fix: https://github.com/redis/redis/pull/9733
         if command in ("EVAL", "EVALSHA"):
             # command syntax: EVAL "script body" num_keys ...
-            if len(args) <= 2:
-                raise RedisClusterException(f"Invalid args in command: {args}")
-            num_actual_keys = args[2]
-            eval_keys = args[3 : 3 + num_actual_keys]
+            if len(args) < 2:
+                raise RedisClusterException(
+                    f"Invalid args in command: {command, *args}"
+                )
+            keys = args[2 : 2 + args[1]]
             # if there are 0 keys, that means the script can be run on any node
             # so we can just return a random slot
-            if not eval_keys:
+            if not keys:
                 return random.randrange(0, REDIS_CLUSTER_HASH_SLOTS)
-            keys = eval_keys
         else:
-            keys = await self.commands_parser.get_keys(
-                self.nodes_manager.default_node, *args
-            )
+            keys = await self.commands_parser.get_keys(command, *args)
             if not keys:
                 # FCALL can call a function with 0 keys, that means the function
                 #  can be run on any node so we can just return a random slot
@@ -516,14 +550,10 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
 
         return slots.pop()
 
-    def _is_node_flag(
-        self, target_nodes: Union[List["ClusterNode"], "ClusterNode", str]
-    ) -> bool:
+    def _is_node_flag(self, target_nodes: Any) -> bool:
         return isinstance(target_nodes, str) and target_nodes in self.node_flags
 
-    def _parse_target_nodes(
-        self, target_nodes: Union[List["ClusterNode"], "ClusterNode"]
-    ) -> List["ClusterNode"]:
+    def _parse_target_nodes(self, target_nodes: Any) -> List["ClusterNode"]:
         if isinstance(target_nodes, list):
             nodes = target_nodes
         elif isinstance(target_nodes, ClusterNode):
@@ -533,7 +563,7 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
             # Supports dictionaries of the format {node_name: node}.
             # It enables to execute commands with multi nodes as follows:
             # rc.cluster_save_config(rc.get_primaries())
-            nodes = target_nodes.values()
+            nodes = list(target_nodes.values())
         else:
             raise TypeError(
                 "target_nodes type can be one of the following: "
@@ -543,7 +573,7 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
             )
         return nodes
 
-    async def execute_command(self, *args: Union[KeyT, EncodableT], **kwargs) -> Any:
+    async def execute_command(self, *args: EncodableT, **kwargs: Any) -> Any:
         """
         Execute a raw command on the appropriate cluster node or target_nodes.
 
@@ -562,7 +592,8 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
             can't be mapped to a slot
         """
         command = args[0]
-        target_nodes_specified = target_nodes = exception = None
+        target_nodes = []
+        target_nodes_specified = False
         retry_attempts = self.cluster_error_retry_attempts
 
         passed_targets = kwargs.pop("target_nodes", None)
@@ -571,7 +602,7 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
             target_nodes_specified = True
             retry_attempts = 1
 
-        for _ in range(0, retry_attempts):
+        for _ in range(retry_attempts):
             if self._initialize:
                 await self.initialize()
             try:
@@ -622,9 +653,10 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
         raise exception
 
     async def _execute_command(
-        self, target_node: "ClusterNode", *args: Union[KeyT, EncodableT], **kwargs
+        self, target_node: "ClusterNode", *args: Union[KeyT, EncodableT], **kwargs: Any
     ) -> Any:
-        redirect_addr = asking = moved = None
+        asking = moved = False
+        redirect_addr = None
         ttl = self.RedisClusterRequestTTL
         connection_error_retry_counter = 0
 
@@ -696,6 +728,24 @@ class RedisCluster(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommand
 
         raise ClusterError("TTL exhausted.")
 
+    def pipeline(
+        self, transaction: Optional[Any] = None, shard_hint: Optional[Any] = None
+    ) -> "ClusterPipeline":
+        """
+        Create & return a new :class:`~.ClusterPipeline` object.
+
+        Cluster implementation of pipeline does not support transaction or shard_hint.
+
+        :raises RedisClusterException: if transaction or shard_hint are truthy values
+        """
+        if shard_hint:
+            raise RedisClusterException("shard_hint is deprecated in cluster mode")
+
+        if transaction:
+            raise RedisClusterException("transaction is deprecated in cluster mode")
+
+        return ClusterPipeline(self)
+
 
 class ClusterNode:
     """
@@ -706,6 +756,7 @@ class ClusterNode:
     """
 
     __slots__ = (
+        "_command_stack",
         "_connections",
         "_free",
         "connection_class",
@@ -723,10 +774,10 @@ class ClusterNode:
         host: str,
         port: int,
         server_type: Optional[str] = None,
-        max_connections: int = 2 ** 31,
+        max_connections: int = 2**31,
         connection_class: Type[Connection] = Connection,
-        response_callbacks: Dict = None,
-        **connection_kwargs,
+        response_callbacks: Dict[str, Any] = RedisCluster.RESPONSE_CALLBACKS,
+        **connection_kwargs: Any,
     ) -> None:
         if host == "localhost":
             host = socket.gethostbyname(host)
@@ -743,8 +794,9 @@ class ClusterNode:
         self.connection_kwargs = connection_kwargs
         self.response_callbacks = response_callbacks
 
-        self._connections = []
-        self._free = collections.deque(maxlen=self.max_connections)
+        self._connections: List[Connection] = []
+        self._free: Deque[Connection] = collections.deque(maxlen=self.max_connections)
+        self._command_stack: List["PipelineCommand"] = []
 
     def __repr__(self) -> str:
         return (
@@ -752,15 +804,15 @@ class ClusterNode:
             f"name={self.name}, server_type={self.server_type}]"
         )
 
-    def __eq__(self, obj: "ClusterNode") -> bool:
+    def __eq__(self, obj: Any) -> bool:
         return isinstance(obj, ClusterNode) and obj.name == self.name
 
     _DEL_MESSAGE = "Unclosed ClusterNode object"
 
-    def __del__(self, _warnings=warnings):
+    def __del__(self) -> None:
         for connection in self._connections:
             if connection.is_connected:
-                _warnings.warn(
+                warnings.warn(
                     f"{self._DEL_MESSAGE} {self!r}", ResourceWarning, source=self
                 )
                 try:
@@ -783,27 +835,26 @@ class ClusterNode:
         if exc:
             raise exc
 
-    async def execute_command(self, *args, **kwargs) -> Any:
-        # Acquire connection
-        connection = None
+    def acquire_connection(self) -> Connection:
         if self._free:
             for _ in range(len(self._free)):
                 connection = self._free.popleft()
                 if connection.is_connected:
-                    break
+                    return connection
                 self._free.append(connection)
-            else:
-                connection = self._free.popleft()
-        else:
-            if len(self._connections) < self.max_connections:
-                connection = self.connection_class(**self.connection_kwargs)
-                self._connections.append(connection)
-            else:
-                raise ConnectionError("Too many connections")
 
-        # Execute command
-        command = connection.pack_command(*args)
-        await connection.send_packed_command(command, False)
+            return self._free.popleft()
+
+        if len(self._connections) < self.max_connections:
+            connection = self.connection_class(**self.connection_kwargs)
+            self._connections.append(connection)
+            return connection
+
+        raise ConnectionError("Too many connections")
+
+    async def parse_response(
+        self, connection: Connection, command: str, **kwargs: Any
+    ) -> Any:
         try:
             if NEVER_DECODE in kwargs:
                 response = await connection.read_response_without_lock(
@@ -815,25 +866,61 @@ class ClusterNode:
             if EMPTY_RESPONSE in kwargs:
                 return kwargs[EMPTY_RESPONSE]
             raise
+
+        # Return response
+        if command in self.response_callbacks:
+            return self.response_callbacks[command](response, **kwargs)
+
+        return response
+
+    async def execute_command(self, *args: Any, **kwargs: Any) -> Any:
+        # Acquire connection
+        connection = self.acquire_connection()
+
+        # Execute command
+        await connection.send_packed_command(connection.pack_command(*args), False)
+
+        # Read response
+        try:
+            return await self.parse_response(connection, args[0], **kwargs)
         finally:
             # Release connection
             self._free.append(connection)
 
-        # Return response
-        try:
-            return self.response_callbacks[args[0]](response, **kwargs)
-        except KeyError:
-            return response
+    async def execute_pipeline(self) -> bool:
+        # Acquire connection
+        connection = self.acquire_connection()
+
+        # Execute command
+        await connection.send_packed_command(
+            connection.pack_commands(cmd.args for cmd in self._command_stack), False
+        )
+
+        # Read responses
+        ret = False
+        for cmd in self._command_stack:
+            try:
+                cmd.result = await self.parse_response(
+                    connection, cmd.args[0], **cmd.kwargs
+                )
+            except Exception as e:
+                cmd.result = e
+                ret = True
+
+        # Release connection
+        self._free.append(connection)
+
+        return ret
 
 
 class NodesManager:
     __slots__ = (
         "_moved_exception",
-        "_require_full_coverage",
         "connection_kwargs",
         "default_node",
         "nodes_cache",
         "read_load_balancer",
+        "require_full_coverage",
         "slots_cache",
         "startup_nodes",
     )
@@ -842,23 +929,24 @@ class NodesManager:
         self,
         startup_nodes: List["ClusterNode"],
         require_full_coverage: bool = False,
-        **kwargs,
+        **kwargs: Any,
     ) -> None:
-        self.nodes_cache = {}
-        self.slots_cache = {}
         self.startup_nodes = {node.name: node for node in startup_nodes}
-        self.default_node = None
-        self._require_full_coverage = require_full_coverage
-        self._moved_exception = None
+        self.require_full_coverage = require_full_coverage
         self.connection_kwargs = kwargs
+
+        self.default_node: "ClusterNode" = None
+        self.nodes_cache: Dict[str, "ClusterNode"] = {}
+        self.slots_cache: Dict[int, List["ClusterNode"]] = {}
         self.read_load_balancer = LoadBalancer()
+        self._moved_exception: MovedError = None
 
     def get_node(
         self,
         host: Optional[str] = None,
         port: Optional[int] = None,
         node_name: Optional[str] = None,
-    ) -> "ClusterNode":
+    ) -> Optional["ClusterNode"]:
         if host and port:
             # the user passed host and port
             if host == "localhost":
@@ -877,20 +965,18 @@ class NodesManager:
         self,
         old: Dict[str, "ClusterNode"],
         new: Dict[str, "ClusterNode"],
-        remove_old=False,
+        remove_old: bool = False,
     ) -> None:
-        tasks = []
         if remove_old:
-            tasks = [
-                asyncio.ensure_future(node.disconnect())
-                for name, node in old.items()
-                if name not in new
-            ]
+            for name in list(old.keys()):
+                if name not in new:
+                    asyncio.ensure_future(old.pop(name).disconnect())
+
         for name, node in new.items():
             if name in old:
                 if old[name] is node:
                     continue
-                tasks.append(asyncio.ensure_future(old[name].disconnect()))
+                asyncio.ensure_future(old[name].disconnect())
             old[name] = node
 
     def _update_moved_slots(self) -> None:
@@ -949,7 +1035,7 @@ class NodesManager:
         except (IndexError, TypeError):
             raise SlotNotCoveredError(
                 f'Slot "{slot}" not covered by the cluster. '
-                f'"require_full_coverage={self._require_full_coverage}"'
+                f'"require_full_coverage={self.require_full_coverage}"'
             )
 
     def get_nodes_by_server_type(self, server_type: str) -> List["ClusterNode"]:
@@ -961,8 +1047,8 @@ class NodesManager:
 
     async def initialize(self) -> None:
         self.read_load_balancer.reset()
-        tmp_nodes_cache = {}
-        tmp_slots = {}
+        tmp_nodes_cache: Dict[str, "ClusterNode"] = {}
+        tmp_slots: Dict[int, List["ClusterNode"]] = {}
         disagreements = []
         startup_nodes_reachable = False
         fully_covered = False
@@ -975,9 +1061,7 @@ class NodesManager:
                     raise RedisClusterException(
                         "Cluster mode is not enabled on this node"
                     )
-                cluster_slots = str_if_bytes(
-                    await startup_node.execute_command("CLUSTER SLOTS")
-                )
+                cluster_slots = await startup_node.execute_command("CLUSTER SLOTS")
                 startup_nodes_reachable = True
             except (ConnectionError, TimeoutError):
                 continue
@@ -1069,7 +1153,7 @@ class NodesManager:
 
             # Validate if all slots are covered or if we should try next startup node
             fully_covered = True
-            for i in range(0, REDIS_CLUSTER_HASH_SLOTS):
+            for i in range(REDIS_CLUSTER_HASH_SLOTS):
                 if i not in tmp_slots:
                     fully_covered = False
                     break
@@ -1083,7 +1167,7 @@ class NodesManager:
             )
 
         # Check if the slots are not fully covered
-        if not fully_covered and self._require_full_coverage:
+        if not fully_covered and self.require_full_coverage:
             # Despite the requirement that the slots be covered, there
             # isn't a full coverage
             raise RedisClusterException(
@@ -1111,3 +1195,242 @@ class NodesManager:
                 for node in getattr(self, attr).values()
             )
         )
+
+
+class ClusterPipeline(AbstractRedis, AbstractRedisCluster, AsyncRedisClusterCommands):
+    """
+    Create a new ClusterPipeline object.
+
+    Usage::
+
+        result = await (
+            rc.pipeline()
+            .set("A", 1)
+            .get("A")
+            .hset("K", "F", "V")
+            .hgetall("K")
+            .mset_nonatomic({"A": 2, "B": 3})
+            .get("A")
+            .get("B")
+            .delete("A", "B", "K")
+            .execute()
+        )
+        # result = [True, "1", 1, {"F": "V"}, True, True, "2", "3", 1, 1, 1]
+
+    Note: For commands `DELETE`, `EXISTS`, `TOUCH`, `UNLINK`, `mset_nonatomic`, which
+    are split across multiple nodes, you'll get multiple results for them in the array.
+
+    Retryable errors:
+        - :class:`~.ClusterDownError`
+        - :class:`~.ConnectionError`
+        - :class:`~.TimeoutError`
+
+    Redirection errors:
+        - :class:`~.TryAgainError`
+        - :class:`~.MovedError`
+        - :class:`~.AskError`
+
+    :param client:
+        | Existing :class:`~.RedisCluster` client
+    """
+
+    __slots__ = ("_command_stack", "_client")
+
+    def __init__(self, client: RedisCluster) -> None:
+        self._client = client
+
+        self._command_stack: List["PipelineCommand"] = []
+
+    async def initialize(self) -> "ClusterPipeline":
+        if self._client._initialize:
+            await self._client.initialize()
+        self._command_stack = []
+        return self
+
+    async def __aenter__(self) -> "ClusterPipeline":
+        return await self.initialize()
+
+    async def __aexit__(self, exc_type: None, exc_value: None, traceback: None) -> None:
+        self._command_stack = []
+
+    def __await__(self) -> Generator[Any, None, "ClusterPipeline"]:
+        return self.initialize().__await__()
+
+    def __enter__(self) -> "ClusterPipeline":
+        self._command_stack = []
+        return self
+
+    def __exit__(self, exc_type: None, exc_value: None, traceback: None) -> None:
+        self._command_stack = []
+
+    def __bool__(self) -> bool:
+        return bool(self._command_stack)
+
+    def __len__(self) -> int:
+        return len(self._command_stack)
+
+    def execute_command(
+        self, *args: Union[KeyT, EncodableT], **kwargs: Any
+    ) -> "ClusterPipeline":
+        """
+        Append a raw command to the pipeline.
+
+        :param args:
+            | Raw command args
+        :param kwargs:
+
+            - target_nodes: :attr:`NODE_FLAGS` or :class:`~.ClusterNode`
+              or List[:class:`~.ClusterNode`] or Dict[Any, :class:`~.ClusterNode`]
+            - Rest of the kwargs are passed to the Redis connection
+        """
+        self._command_stack.append(
+            PipelineCommand(len(self._command_stack), *args, **kwargs)
+        )
+        return self
+
+    async def execute(
+        self, raise_on_error: bool = True, allow_redirections: bool = True
+    ) -> List[Any]:
+        """
+        Execute the pipeline.
+
+        It will retry the commands as specified by :attr:`cluster_error_retry_attempts`
+        & then raise an exception.
+
+        :param raise_on_error:
+            | Raise the first error if there are any errors
+        :param allow_redirections:
+            | Whether to retry each failed command individually in case of redirection
+              errors
+
+        :raises RedisClusterException: if target_nodes is not provided & the command
+            can't be mapped to a slot
+        """
+        if not self._command_stack:
+            return []
+
+        try:
+            for _ in range(self._client.cluster_error_retry_attempts):
+                if self._client._initialize:
+                    await self._client.initialize()
+
+                try:
+                    return await self._execute(
+                        self._client,
+                        self._command_stack,
+                        raise_on_error=raise_on_error,
+                        allow_redirections=allow_redirections,
+                    )
+                except BaseException as e:
+                    if type(e) in self.__class__.ERRORS_ALLOW_RETRY:
+                        # Try again with the new cluster setup.
+                        exception = e
+                        await self._client.close()
+                        await asyncio.sleep(0.25)
+                    else:
+                        # All other errors should be raised.
+                        raise e
+
+            # If it fails the configured number of times then raise an exception
+            raise exception
+        finally:
+            self._command_stack = []
+
+    async def _execute(
+        self,
+        client: "RedisCluster",
+        stack: List["PipelineCommand"],
+        raise_on_error: bool = True,
+        allow_redirections: bool = True,
+    ) -> List[Any]:
+        todo = [
+            cmd for cmd in stack if not cmd.result or isinstance(cmd.result, Exception)
+        ]
+
+        nodes = {}
+        for cmd in todo:
+            target_nodes = await client._determine_nodes(*cmd.args)
+            if not target_nodes:
+                raise RedisClusterException(
+                    f"No targets were found to execute {cmd.args} command on"
+                )
+            if len(target_nodes) > 1:
+                raise RedisClusterException(f"Too many targets for command {cmd.args}")
+
+            node = target_nodes[0]
+            if node.name not in nodes:
+                nodes[node.name] = node
+                node._command_stack = []
+            node._command_stack.append(cmd)
+
+        errors = await asyncio.gather(
+            *(asyncio.ensure_future(node.execute_pipeline()) for node in nodes.values())
+        )
+
+        if any(errors):
+            if allow_redirections:
+                # send each errored command individually
+                for cmd in todo:
+                    if isinstance(cmd.result, (TryAgainError, MovedError, AskError)):
+                        try:
+                            cmd.result = await client.execute_command(
+                                *cmd.args, **cmd.kwargs
+                            )
+                        except Exception as e:
+                            cmd.result = e
+
+            if raise_on_error:
+                for cmd in todo:
+                    result = cmd.result
+                    if isinstance(result, Exception):
+                        command = " ".join(map(safe_str, cmd.args))
+                        msg = (
+                            f"Command # {cmd.position + 1} ({command}) of pipeline "
+                            f"caused error: {result.args}"
+                        )
+                        result.args = (msg,) + result.args[1:]
+                        raise result
+
+        return [cmd.result for cmd in stack]
+
+    def _split_command_across_slots(
+        self, command: str, *keys: KeyT
+    ) -> "ClusterPipeline":
+        for slot_keys in self._client._partition_keys_by_slot(keys).values():
+            self.execute_command(command, *slot_keys)
+
+        return self
+
+    def mset_nonatomic(
+        self, mapping: Mapping[AnyKeyT, EncodableT]
+    ) -> "ClusterPipeline":
+        encoder = self._client.encoder
+
+        slots_pairs = {}
+        for pair in mapping.items():
+            slot = key_slot(encoder.encode(pair[0]))
+            slots_pairs.setdefault(slot, []).extend(pair)
+
+        for pairs in slots_pairs.values():
+            self.execute_command("MSET", *pairs)
+
+        return self
+
+
+for command in PIPELINE_BLOCKED_COMMANDS:
+    command = command.replace(" ", "_").lower()
+    if command == "mset_nonatomic":
+        continue
+
+    setattr(ClusterPipeline, command, block_pipeline_command(command))
+
+
+class PipelineCommand:
+    def __init__(self, position: int, *args: Any, **kwargs: Any) -> None:
+        self.args = args
+        self.kwargs = kwargs
+        self.position = position
+        self.result: Union[Any, Exception] = None
+
+    def __repr__(self) -> str:
+        return f"[{self.position}] {self.args} ({self.kwargs})"
